import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
import torch.optim as optim
import cv2
from PIL import Image
import numpy as np

# 获取数据集
data_path = 'data'
transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])
cifar_train = torchvision.datasets.CIFAR10(root=data_path, train=True,download=True, transform=transform)
cifar_test  = torchvision.datasets.CIFAR10(root=data_path, train=False,download=True, transform=transform)

# 加载数据集
batch_size = 32
trainloader = torch.utils.data.DataLoader(cifar_train, batch_size=batch_size, shuffle=True)
testloader = torch.utils.data.DataLoader(cifar_test, batch_size=batch_size, shuffle=False)


# 搭建模型
class CNNNet(nn.Module):
    def __init__(self):
        super(CNNNet, self).__init__()
        # Conv2d的第一个参数是输入的channel数量，第二个是输出的channel数量，第三个是kernel size
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.conv2 = nn.Conv2d(6, 16, 5)
        # 由于上一层有16个channel输出，每个feature map大小为5*5，所以全连接层的输入是16*5*5
        self.fc1 = nn.Linear(16*5*5, 120)
        self.fc2 = nn.Linear(120, 84)
        # 最终有10类，所以最后一个全连接层输出数量是10
        self.fc3 = nn.Linear(84, 10)
        self.pool = nn.MaxPool2d(2, 2)
    # forward这个函数定义了前向传播的运算，只需要像写普通的python算数运算那样就可以了
    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = self.pool(x)
        x = F.relu(self.conv2(x))
        x = self.pool(x)
        # 下面这步把二维特征图变为一维，这样全连接层才能处理
        x = x.view(-1, 16*5*5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

# 定义交叉熵损失函数和初始化模型
model = CNNNet()
criterion = nn.CrossEntropyLoss()
optimizer = optim.SGD(model.parameters(), lr=0.01) #这里设置优化器和学习率

# ## 模型训练

epochs = 10 #设置训epoch
save_path = "cifar10_epoch{}.pth.tar".format(epochs)
print("Start Training...")
loss_list = []#记录每个epoch的损失值
for epoch in range(epochs):
    # 变量来记录每epoch的平均loss,acc
    losss = 0.0
    accc = 0.0
    for i, data in enumerate(trainloader):
        inputs, labels = data
        optimizer.zero_grad()
        outputs = model(inputs)
        loss = criterion(outputs, labels)
        _, pred = torch.max(outputs.view(-1, 10), 1)
        acc = float((pred == labels).long().sum()) / labels.size(0)
        loss.backward()
        optimizer.step()
        losss += loss.item()
        accc += acc
        # print('[Epoch %d] loss: %.3f acc:  %.3f ' % (epoch + 1, losss / (i+1), accc / (i+1)))
    loss_list.append(losss/(i+1))
# 保存模型参数
torch.save(model.state_dict(), save_path)
print("Done Training!")

# ## 在测试集上准确率
model_weight = torch.load(save_path)
model.load_state_dict(model_weight)
# 构造测试的dataloader
dataiter = iter(testloader)
# 预测正确的数量和总数量
correct = 0
total = 0
with torch.no_grad():
    for images, labels in dataiter:
        outputs = model(images)
        # 用最大概率的一项作为预测分类
        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()
print('Accuracy of 10000 test images: %d %%' % (100 * correct / total))

# ## 模型展示 

idx = 15 #随机取测试集一张图片
data = cifar_test.data[idx]#获取测试集的一张图片
t = cifar_test.targets[idx]#获取标签
t = cifar_test.classes[t]
print("标签是：{}".format(t))

data = transform(data).unsqueeze(0)#对原始数据转化
outputs = model(data)#进行模型预测
_, pred = torch.max(outputs.data, 1)#取预测概率最大值
pred = cifar_test.classes[pred.item()]#查看预测的类别名称
print("预测结果是：{}".format(pred))