import re
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
import torch
from torch import nn, optim
from torch.utils.data import DataLoader, Dataset
from sklearn.metrics import classification_report
import optuna

# 加载数据集
file_path = 'results/train_data.csv'
data = pd.read_csv(file_path)

# 合并相关文本列进行分析
data['text'] = data['title'].fillna('') + " " + data['company_profile'].fillna('') + " " + data['description'] + " " + data['requirements'].fillna('') + " " + data['benefits'].fillna('')

# 清理文本数据
def clean_text(text):
    text = re.sub(r'[^a-zA-Z0-9\s]', '', text, re.I|re.A)
    text = text.lower()
    text = text.strip()
    return text

data['text'] = data['text'].apply(clean_text)

# 对文本数据进行分词
from sklearn.feature_extraction.text import CountVectorizer
vectorizer = CountVectorizer()
X = vectorizer.fit_transform(data['text']).toarray()
vocab_size = len(vectorizer.vocabulary_)

# 提取目标变量
y = data['fraudulent'].values

# 将数据分为训练集和验证集
X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=0.2, random_state=42)

# 定义数据集类
class JobDataset(Dataset):
    def __init__(self, texts, labels):
        self.texts = texts
        self.labels = labels

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, idx):
        return torch.tensor(self.texts[idx], dtype=torch.float32), torch.tensor(self.labels[idx], dtype=torch.float32)

# 创建训练和验证数据集
train_dataset = JobDataset(X_train, y_train)
val_dataset = JobDataset(X_val, y_val)

# 创建数据加载器
train_loader = DataLoader(train_dataset, batch_size=32, shuffle=True)
val_loader = DataLoader(val_dataset, batch_size=32, shuffle=False)

# 定义LSTM分类器
class LSTMClassifier(nn.Module):
    def __init__(self, vocab_size, embedding_dim, hidden_dim):
        super(LSTMClassifier, self).__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        self.lstm = nn.LSTM(embedding_dim, hidden_dim, batch_first=True)
        self.fc = nn.Linear(hidden_dim, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.embedding(x)
        x, (h, c) = self.lstm(x)
        x = self.fc(x[:, -1, :])
        x = self.sigmoid(x)
        return x

# 训练模型
def train_model(model, criterion, optimizer, train_loader, val_loader, epochs=10):
    for epoch in range(epochs):
        model.train()
        for texts, labels in train_loader:
            # texts, labels=texts.cuda(), labels.cuda()
            optimizer.zero_grad()
            outputs = model(texts.long())
            loss = criterion(outputs, labels.unsqueeze(1))
            loss.backward()
            optimizer.step()

        model.eval()
        val_loss = 0
        with torch.no_grad():
            for texts, labels in val_loader:
                outputs = model(texts.long())
                loss = criterion(outputs, labels.unsqueeze(1))
                val_loss += loss.item()
        print(f'Epoch {epoch+1}, Validation Loss: {val_loss / len(val_loader)}')

# 定义Optuna目标函数
def objective(trial):
    embedding_dim = trial.suggest_int('embedding_dim', 2, 8)
    hidden_dim = trial.suggest_int('hidden_dim', 4, 16)
    # embedding_dim = trial.suggest_int('embedding_dim', 32, 128)
    # hidden_dim = trial.suggest_int('hidden_dim', 64, 256)
    lr = trial.suggest_loguniform('lr', 1e-5, 1e-1)

    model = LSTMClassifier(vocab_size=vocab_size, embedding_dim=embedding_dim, hidden_dim=hidden_dim)
    model = model
    criterion = nn.BCELoss()
    optimizer = optim.Adam(model.parameters(), lr=lr)

    train_model(model, criterion, optimizer, train_loader, val_loader, epochs=10)

    val_loss = 0
    model.eval()
    with torch.no_grad():
        for texts, labels in val_loader:
            outputs = model(texts.long())
            loss = criterion(outputs, labels.unsqueeze(1))
            val_loss += loss.item()
    
    return val_loss / len(val_loader)

# 创建Optuna学习对象
study = optuna.create_study(direction='minimize')
study.optimize(objective, n_trials=50)

print('Best trial:')
trial = study.best_trial

print(f'  Val Loss: {trial.value}')
print('  Params: ')
for key, value in trial.params.items():
    print(f'    {key}: {value}')

# 使用最佳超参数训练最终模型
best_params = trial.params
final_model = LSTMClassifier(vocab_size=vocab_size, embedding_dim=best_params['embedding_dim'], hidden_dim=best_params['hidden_dim'])
final_optimizer = optim.Adam(final_model.parameters(), lr=best_params['lr'])

train_model(final_model, nn.BCELoss(), final_optimizer, train_loader, val_loader, epochs=10)

# 保存模型
model_save_path = 'model/model.pt'
torch.save(final_model, model_save_path)
print(f"Model saved to {model_save_path}")


# ********************************下面为预测结果，可以不使用***********************************************



# 评估模型
def evaluate_model(model, data_loader):
    model.eval()
    y_true = []
    y_pred = []
    with torch.no_grad():
        for texts, labels in data_loader:
            outputs = model(texts.long())
            y_true.extend(labels.tolist())
            y_pred.extend(outputs.squeeze().tolist())
    return y_true, y_pred

y_true, y_pred = evaluate_model(final_model, val_loader)
y_pred = [1 if p > 0.5 else 0 for p in y_pred]

print(classification_report(y_true, y_pred))
