import re
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Embedding, LSTM, Dense, Dropout
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import EarlyStopping
from sklearn.metrics import classification_report
from keras_tuner import RandomSearch, HyperModel

# 加载数据集
file_path = 'results/train_data.csv'
data = pd.read_csv(file_path)

# 合并相关数据
data['text'] = data['title'].fillna('') + " " + data['company_profile'].fillna('') + " " + data['description'] + " " + data['requirements'].fillna('') + " " + data['benefits'].fillna('')

# 清晰数据
def clean_text(text):
    text = re.sub(r'[^a-zA-Z0-9\s]', '', text, re.I|re.A)
    text = text.lower()
    text = text.strip()
    return text

data['text'] = data['text'].apply(clean_text)

# 分词
tokenizer = Tokenizer()
tokenizer.fit_on_texts(data['text'].values)
word_index = tokenizer.word_index

# 转换向量
sequences = tokenizer.texts_to_sequences(data['text'].values)

# 填充序列
max_length = max([len(x) for x in sequences])
X = pad_sequences(sequences, maxlen=max_length)

# 提取目标变量
y = data['fraudulent'].values

# 将数据拆分为训练集和验证集
X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=0.2, random_state=42)

# 定义超参数
class MyHyperModel(HyperModel):
    def build(self, hp):
        model = Sequential()
        model.add(Embedding(input_dim=len(word_index) + 1,
                            output_dim=hp.Choice('output_dim', values=[64, 128, 256]),
                            input_length=max_length))
        model.add(LSTM(units=hp.Choice('units', values=[64, 128, 256]),
                       dropout=hp.Float('dropout', min_value=0.2, max_value=0.5, step=0.1),
                       recurrent_dropout=hp.Float('recurrent_dropout', min_value=0.2, max_value=0.5, step=0.1)))
        model.add(Dense(units=1, activation='sigmoid'))
        
        model.compile(optimizer=Adam(learning_rate=hp.Choice('learning_rate', values=[1e-2, 1e-3, 1e-4])),
                      loss='binary_crossentropy',
                      metrics=['accuracy'])
        return model

# 初始化
tuner = RandomSearch(
    MyHyperModel(),
    objective='val_accuracy',
    max_trials=5,
    executions_per_trial=3,
    directory='tuner_dir',
    project_name='fraud_detection'
)

# 定义 early stopping
early_stopping = EarlyStopping(monitor='val_loss', patience=3, restore_best_weights=True)

# 最佳参数查询
tuner.search(X_train, y_train, validation_data=(X_val, y_val), epochs=10, batch_size=64, callbacks=[early_stopping])

# 检索最佳模型
best_model = tuner.get_best_models(num_models=1)[0]

# 保存最佳模型
best_model.save('model/tensor/best_model.h5')


# ********************************下面为预测结果，可以不使用***********************************************


# 评估模型
y_pred = (best_model.predict(X_val) > 0.5).astype("int32")
report = classification_report(y_val, y_pred)
print(report)

# 加载数据集并预测
new_data_path = 'results/test_data.csv'
new_data = pd.read_csv(new_data_path)

# 合并相关信息
new_data['text'] = new_data['title'] + " " + new_data['company_profile'].fillna('') + " " + new_data['description'] + " " + new_data['requirements'].fillna('') + " " + new_data['benefits'].fillna('')

# 清理数据
new_data['text'] = new_data['text'].apply(clean_text)

# 转换序列
sequences_new = tokenizer.texts_to_sequences(new_data['text'].values)

# 填充序列
X_new = pad_sequences(sequences_new, maxlen=max_length)

# 使用最佳模型预测
predictions = (best_model.predict(X_new) > 0.5).astype("int32")

# 赋值生成结果
new_data['fraudulent'] = predictions

# 保存预测
new_data[['job_id', 'fraudulent']].to_csv("results/results.csv", index=False)

print("Results saved")
