import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, Dataset
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from nltk.tokenize import word_tokenize
import torch.nn as nn
import torch.optim as optim
df = pd.read_csv('dataset/fjp.csv')
df['text'] = df['title'].fillna('') + ' ' + df['company_profile'].fillna('') + " " + df['description'] + " " + df['requirements'].fillna('') + " " + df['benefits'].fillna('')


class TextDataset(Dataset):
    def __init__(self, texts, labels, vocab, max_length):
        self.texts = texts
        self.labels = labels
        self.vocab = vocab
        self.max_length = max_length

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, idx):
        text = self.texts[idx]
        label = self.labels[idx]

        # tokenize
        tokens = word_tokenize(text)
        indices = [self.vocab.get(token, self.vocab['<UNK>']) for token in tokens]
        if len(indices) < self.max_length:
            indices += [self.vocab['<PAD>']]* (self.max_length - len(indices))
        else:
            indices = indices[:self.max_length]

        return torch.tensor(indices, dtype=torch.long), torch.tensor(label, dtype=torch.long)


def build_vocab(texts, min_freq =1):
    vocab = {'<PAD>':0, '<UNK>':1}
    word_freq = {}
    for text in texts:
        tokens = word_tokenize(text)
        for token in tokens:
            if token not in word_freq:
                word_freq[token] = 1
            else:
                word_freq[token] += 1
    for word, freq in word_freq.items():
        if freq > min_freq:
            vocab[word] = len(vocab)

    return vocab

texts = df['text'].values
labels =  LabelEncoder().fit_transform(df['fraudulent'].values)

# 划分训练集和测试集
train_texts, test_texts, train_labels, test_labels = train_test_split(texts, labels, test_size=0.2, random_state=42)
vocab = build_vocab(train_texts)
max_length = 100

train_dataset = TextDataset(train_texts, train_labels, vocab, max_length)
test_dataset = TextDataset(test_texts, test_labels, vocab, max_length)

train_loader = DataLoader(train_dataset, batch_size=32, shuffle=True)
test_loader = DataLoader(test_dataset, batch_size=32, shuffle=False)

class LSTMClassifier(nn.Module):
    def __init__(self,vocab_size, embedding_dim ,hidden_dim, output_dim, n_layers, dropout):
        super(LSTMClassifier,self).__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        self.lstm = nn.LSTM(embedding_dim, hidden_dim, n_layers, batch_first=True, dropout=dropout)
        self.fc = nn.Linear(hidden_dim, output_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self,x):
        embedded = self.dropout(self.embedding(x))
        lstm_out, (hidden, cell) = self.lstm(embedded)
        hidden = self.dropout(hidden[-1])
        output = self.fc(hidden)
        return output

vocab_size = len(vocab)
embedding_dim = 100
hidden_dim = 128
output_dim = len(np.unique(labels))
n_layers = 2
dropout = 0.3


model = LSTMClassifier(vocab_size, embedding_dim, hidden_dim, output_dim, n_layers, dropout)
criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(),lr=0.001)

n_epochs = 1
for epoch in range(n_epochs):
    model.train()
    for texts, labels in train_loader:
        optimizer.zero_grad()
        outputs = model(texts)
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()

    model.eval()
    all_preds = []
    all_labels = []
    for texts, labels in test_loader:
        outputs = model(texts)
        preds = torch.argmax(outputs, dim=1)
        all_preds.extend(preds.numpy())
        all_labels.extend(labels.numpy())

    acc = accuracy_score(all_labels, all_preds)

    print(acc)

