import pandas as pd
import torch
from transformers import LlamaTokenizer, LlamaForSequenceClassification
from sklearn.metrics import accuracy_score
from torch.utils.data import DataLoader, Dataset

# 加载数据
validation_file_path = 'task2/Sentiment_Analysis/dataset/validation_data.csv'
validation_data = pd.read_csv(validation_file_path)

# 标签
label_map = {'positive': 1, 'negative': 0}
validation_data['label'] = validation_data['sentiment'].map(label_map)

# 加载分词器
tokenizer = LlamaTokenizer.from_pretrained('meta-llama/Meta-Llama-3-8B')

# 对输入进行分词
val_encodings = tokenizer(validation_data['review'].tolist(), truncation=True, padding=True)

# 创建数据类
class SentimentDataset(Dataset):
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item['labels'] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)

# 加载验证集
val_dataset = SentimentDataset(val_encodings, validation_data['label'].tolist())

# 加载模型
model = LlamaForSequenceClassification.from_pretrained('task2/Sentiment_Analysis/result/best_model.pt')

# 创建DataLoader
val_loader = DataLoader(val_dataset, batch_size=8)

# 开始评估
model.eval()
predictions, true_labels = [], []

# 遍历数据
for batch in val_loader:
    with torch.no_grad():   # 关闭自动求导（gradient computation），以节省内存和计算资源。
        inputs = {key: val.to('cuda') for key, val in batch.items() if key != 'labels'} # 将数据挪到GPU
        labels = batch['labels'].to('cuda') # 将数据挪到GPU
        outputs = model(**inputs) # 输入数据，由模型来解包并输出
        logits = outputs.logits # 获取当前批次的logits分数
        preds = torch.argmax(logits, dim=-1) # 计算每个样本数据的类别，进行预测
        predictions.extend(preds.cpu().numpy()) # 转换为numpy，并添加到列表里面
        true_labels.extend(labels.cpu().numpy()) # 真实数据添加到列表

# 计算精准度
accuracy = accuracy_score(true_labels, predictions)
print(f'Validation Accuracy: {accuracy:.4f}')
