import torch
import torch.nn as nn
import torch.nn.functional as F
class depthwise_separable_conv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1,padding=1, dilation=1):
        super(depthwise_separable_conv, self).__init__()
        self.depthwise = nn.Conv2d(in_channels, in_channels, kernel_size, stride, padding, dilation, groups=in_channels,bias=False)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        
    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out
 
class XceptionABlock(nn.Module):
    """
    Base Block for XceptionA mentioned in DFANet paper.
    """
    def __init__(self, in_channels, out_channels, stride=1):
        super(XceptionABlock, self).__init__()
        self.conv1 = nn.Sequential(
            depthwise_separable_conv(in_channels, out_channels //4, stride=stride),
            nn.BatchNorm2d(out_channels // 4),
            nn.ReLU(),
        )
        self.conv2 = nn.Sequential(
            depthwise_separable_conv(out_channels //4, out_channels //4),
            nn.BatchNorm2d(out_channels // 4),
            nn.ReLU(),
        )
        self.conv3 = nn.Sequential(
            depthwise_separable_conv(out_channels //4, out_channels),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
        )
 
        self.skip = nn.Conv2d(in_channels, out_channels, 1, stride=stride, bias=False)
 
    def forward(self, x):
        residual = self.conv1(x)
        residual = self.conv2(residual)
        residual = self.conv3(residual)
        identity = self.skip(x)
        
        return residual + identity
        
class enc(nn.Module):
    """
    encoder block
    """
    def __init__(self, in_channels, out_channels, stride=2, num_repeat=3):
        super(enc, self).__init__()
        stacks = [XceptionABlock(in_channels, out_channels, stride=2)]
        for x in range(num_repeat - 1):
            stacks.append(XceptionABlock(out_channels, out_channels))
        self.build = nn.Sequential(*stacks)
        
    def forward(self, x):
        x = self.build(x)
        return x
    
class ChannelAttention(nn.Module):
    """
        channel attention module
    """
    def __init__(self, in_channels, out_channels):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_channels, 1000, bias=False),
            nn.ReLU(),
            nn.Linear(1000, out_channels, bias=False),
            nn.Sigmoid()
        )
        
    def forward(self, x):
        b,c,_,_ = x.size()
        y = self.avg_pool(x).view(b,c)
        y = self.fc(y).view(b,c,1,1)
        return x*y.expand_as(x)
 
class SubBranch(nn.Module):
    """
        create 3 Sub Branches in DFANet
        channel_cfg: the chnnels of each enc stage
        branch_index: the index of each sub branch
    """
    def __init__(self, channel_cfg, branch_index):
        super(SubBranch, self).__init__()
        self.enc2 = enc(channel_cfg[0], 48, num_repeat=3)
        self.enc3 = enc(channel_cfg[1],96,num_repeat=6)
        self.enc4 = enc(channel_cfg[2],192,num_repeat=3)
        self.fc_atten = ChannelAttention(192, 192)
        self.branch_index = branch_index
            
        
    def forward(self,x0,*args):
        out0=self.enc2(x0)
        if self.branch_index in [1,2]:
            out1 = self.enc3(torch.cat([out0,args[0]],1))
            out2 = self.enc4(torch.cat([out1,args[1]],1))
        else:
            out1 = self.enc3(out0)
            out2 = self.enc4(out1)
        out3 = self.fc_atten(out2)
        return [out0, out1, out2, out3]
    
class XceptionA(nn.Module):
    """
    channel_cfg: the all channels in each enc blocks
    """
    def __init__(self, channel_cfg, num_classes=33):
        super(XceptionA, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=3, out_channels=8, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(8),
            nn.ReLU()
        )
        
        self.branch = SubBranch(channel_cfg, branch_index=0)
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        
        self.classifier = nn.Sequential(
            nn.Linear(192, 1000),
            nn.ReLU(),
            nn.Linear(1000, num_classes)
        )
    
    def forward(self, x):
        b,c,_,_ = x.szie()
        x = self.conv1(x)
        _,_,_,x = self.branch(x)
        x = self.avg_pool(x).view(b,-1)
        x = self.classifier(x)
        
        return x
 
class DFA_Encoder(nn.Module):
    def __init__(self, channel_cfg):
        super(DFA_Encoder, self).__init__()
        self.conv1=nn.Sequential(
            nn.Conv2d(in_channels=3,out_channels=8,kernel_size=3,stride=2,padding=1,bias=False),
            nn.BatchNorm2d(num_features=8),
            nn.ReLU()
        )
        self.branch0=SubBranch(channel_cfg[0],branch_index=0)
        self.branch1=SubBranch(channel_cfg[1],branch_index=1)
        self.branch2=SubBranch(channel_cfg[2],branch_index=2)
    
    def forward(self, x):
        x = self.conv1(x)
        x0,x1,x2,x5=self.branch0(x)
        x3=F.interpolate(x5,x0.size()[2:],mode='bilinear',align_corners=True)
        
        x1,x2,x3,x6=self.branch1(torch.cat([x0,x3],1),x1,x2)
        x4=F.interpolate(x6,x1.size()[2:],mode='bilinear',align_corners=True)
        
        x2,x3,x4,x7=self.branch2(torch.cat([x1,x4],1),x2,x3)
 
        return [x0,x1,x2,x5,x6,x7]
    
class DFA_Decoder(nn.Module):
    """
        the DFA decoder 
    """
    def __init__(self, decode_channels, num_classes):
        super(DFA_Decoder, self).__init__()
        
        self.conv0 = nn.Sequential(
            nn.Conv2d(in_channels=48, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=48, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(in_channels=48, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(in_channels=192, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv4 = nn.Sequential(
            nn.Conv2d(in_channels=192, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv5 = nn.Sequential(
            nn.Conv2d(in_channels=192, out_channels=decode_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv_add = nn.Sequential(
            nn.Conv2d(in_channels=decode_channels, out_channels=decode_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(decode_channels),
            nn.ReLU()
        )
        self.conv_cls = nn.Conv2d(in_channels=decode_channels, out_channels=num_classes, kernel_size=3, padding=1, bias=False)
        
    def forward(self, x0, x1, x2, x3, x4, x5):
        x0 = self.conv0(x0)
        x1 = F.interpolate(self.conv1(x1),x0.size()[2:],mode='bilinear',align_corners=True)
        x2 = F.interpolate(self.conv2(x2),x0.size()[2:],mode='bilinear',align_corners=True)
        x3 = F.interpolate(self.conv3(x3),x0.size()[2:],mode='bilinear',align_corners=True)
        x4 = F.interpolate(self.conv5(x4),x0.size()[2:],mode='bilinear',align_corners=True)
        x5 = F.interpolate(self.conv5(x5),x0.size()[2:],mode='bilinear',align_corners=True)
        
        x_shallow = self.conv_add(x0+x1+x2)
        
        x = self.conv_cls(x_shallow+x3+x4+x5)
        x=F.interpolate(x,scale_factor=4,mode='bilinear',align_corners=True)
        return x
    
class DFANet(nn.Module):
    def __init__(self,channel_cfg,decoder_channel,num_classes):
        super(DFANet,self).__init__()
        self.encoder=DFA_Encoder(channel_cfg)
        self.decoder=DFA_Decoder(decoder_channel,num_classes)
 
    def forward(self,x):
        x0,x1,x2,x3,x4,x5=self.encoder(x)
        x=self.decoder(x0,x1,x2,x3,x4,x5)
        return x

#可以用以下代码进行测试
import torch
ch_cfg=[[8,48,96],
        [240,144,288],
        [240,144,288]]
 
device = torch.device("cpu")
model = DFANet(ch_cfg,64,33)
# model = model.to(device)
a = torch.ones([16, 3, 224, 224])
a = a.to(device)
out = model(a)
print(out.shape)



# 导入库
import os
# os.environ['CUDA_VISIBLE_DEVICES'] = '0'
 
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch import optim
from torch.utils.data import Dataset, DataLoader, random_split
from tqdm import tqdm
import warnings
warnings.filterwarnings("ignore")
import os.path as osp
import matplotlib.pyplot as plt
from PIL import Image
import numpy as np
import albumentations as A
from albumentations.pytorch.transforms import ToTensorV2
from torchvision import transforms, models 
import cv2
torch.manual_seed(17)
# 自定义数据集CamVidDataset
# class CamVidDataset(torch.utils.data.Dataset):
#     """CamVid Dataset. Read images, apply augmentation and preprocessing transformations.
    
#     Args:
#         images_dir (str): path to images folder
#         masks_dir (str): path to segmentation masks folder
#         class_values (list): values of classes to extract from segmentation mask
#         augmentation (albumentations.Compose): data transfromation pipeline 
#             (e.g. flip, scale, etc.)
#         preprocessing (albumentations.Compose): data preprocessing 
#             (e.g. noralization, shape manipulation, etc.)
#     """
    
#     def __init__(self, images_dir, masks_dir):
#         self.transform = A.Compose([
#             A.Resize(960, 720),
#             A.HorizontalFlip(),
#             A.VerticalFlip(),
#             A.Normalize(),
#             ToTensorV2(),
#         ]) 
#         self.ids = os.listdir(images_dir)
#         self.mids = os.listdir(masks_dir)
#         self.images_fps = [os.path.join(images_dir, image_id) for image_id in self.ids]
#         self.masks_fps = [os.path.join(masks_dir, image_id) for image_id in self.mids]
#         print(self.images_fps)
#         print(self.masks_fps)
 
    
#     def __getitem__(self, i):
#         # read data
#         image = np.array(Image.open(self.images_fps[i]).convert('RGB'))
#         mask = np.array( Image.open(self.masks_fps[i]).convert('RGB'))
#         image = self.transform(image=image,mask=mask)
        
#         return image['image'], image['mask'][:,:,0]
        
#     def __len__(self):
#         return len(self.ids)
    
    
# # 设置数据集路径
# DATA_DIR = r'dataset' # 根据自己的路径来设置
# x_train_dir = os.path.join(DATA_DIR, 'train_images')
# y_train_dir = os.path.join(DATA_DIR, 'train_labels')
# x_valid_dir = os.path.join(DATA_DIR, 'valid_images')
# y_valid_dir = os.path.join(DATA_DIR, 'valid_labels')
    
# train_dataset = CamVidDataset(
#     x_train_dir, 
#     y_train_dir, 
# )
# val_dataset = CamVidDataset(
#     x_valid_dir, 
#     y_valid_dir, 
# )
 
# train_loader = DataLoader(train_dataset, batch_size=8, shuffle=True,drop_last=True)
# val_loader = DataLoader(val_dataset, batch_size=8, shuffle=True,drop_last=True)




class CityscapesDataset(Dataset):
    def __init__(self, image_paths, label_paths, transform=None):
        self.image_paths = image_paths
        self.label_paths = label_paths
        self.transform = transform

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image = cv2.imread(self.image_paths[idx])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        label = cv2.imread(self.label_paths[idx], cv2.IMREAD_GRAYSCALE)
        
        if self.transform:
            image = self.transform(image)
        
        label = cv2.resize(label, (image.shape[2], image.shape[1]), interpolation=cv2.INTER_NEAREST)
        label = torch.from_numpy(label).long()
        
        # 确保标签在有效范围内
        label[label >= 20] = 19  # 假设19是最高的有效类索引

        return image, label
import glob
# 获取所有图像和标签的路径
image_paths = sorted(glob.glob('dataset/leftImg8bit_trainvaltest/leftImg8bit/train/*_leftImg8bit.png'))
label_paths = sorted(glob.glob('dataset/gtFine/train/*_gtFine_labelIds.png'))

# 确保路径数量一致
assert len(image_paths) == len(label_paths)

# 数据增强和预处理
transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.Resize((256, 512)),
    transforms.ToTensor(),
])

dataset = CityscapesDataset(image_paths, label_paths, transform=transform)
dataloader = DataLoader(dataset, batch_size=4, shuffle=True)

# 验证集加载
val_image_paths = sorted(glob.glob('dataset/leftImg8bit_trainvaltest/leftImg8bit/val/*_leftImg8bit.png'))
val_label_paths = sorted(glob.glob('dataset/gtFine/val/*_gtFine_labelIds.png'))
val_dataset = CityscapesDataset(val_image_paths, val_label_paths, transform=transform)
val_dataloader = DataLoader(val_dataset, batch_size=4, shuffle=False)

train_loader=dataloader
val_loader=val_dataloader


#model.load_state_dict(torch.load(r"checkpoints/resnet50-11ad3fa6.pth"),strict=False)
ch_cfg=[[8,48,96],
        [240,144,288],
        [240,144,288]]
 
# device = torch.device("cuda")
model = DFANet(ch_cfg,64,33)
# model = model.to(device)

# model.classifier[4] = nn.Conv2d(256, 20, kernel_size=(1, 1), stride=(1, 1))  # Cityscapes有19类+背景，共20类
# model = model.cuda()

criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(), lr=0.001)

# 训练模型
num_epochs = 1
for epoch in range(num_epochs):
    model.train()
    running_loss = 0.0
    with tqdm(total=len(dataloader), desc=f"Epoch {epoch+1}/{num_epochs}", unit="batch") as pbar:
        for images, labels in dataloader:
            # images = images.cuda()
            # labels = labels.cuda()

            optimizer.zero_grad()
            print(type(model(images)) )
            outputs = model(images)#['out']
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
            
            running_loss += loss.item()
            pbar.set_postfix(loss=running_loss/(pbar.n + 1))
            pbar.update(1)
    
    print(f"Epoch [{epoch+1}/{num_epochs}], Loss: {running_loss/len(dataloader)}")

    # 验证模型
    model.eval()
    val_loss = 0.0
    with torch.no_grad():
        for images, labels in val_dataloader:
            # images = images.cuda()
            # labels = labels.cuda()
            
            outputs = model(images)#['out']
            loss = criterion(outputs, labels)
            val_loss += loss.item()
    
    print(f"Validation Loss: {val_loss/len(val_dataloader)}")


# 选择一些测试图像进行预测
test_images, test_labels = next(iter(val_dataloader))
# test_images = test_images.cuda()
# test_labels = test_labels.cuda()

# 进行预测
model.eval()
with torch.no_grad():
    predictions = model(test_images)#['out']
predictions = torch.argmax(predictions, dim=1).cpu().numpy()

# 将预测结果转换为类别标签并可视化
for i in range(4):  # 批量大小为4
    plt.figure(figsize=(15, 5))
    
    plt.subplot(1, 3, 1)
    plt.title('Original Image')
    plt.imshow(test_images[i].permute(1, 2, 0).cpu().numpy())
    
    plt.subplot(1, 3, 2)
    plt.title('Ground Truth')
    plt.imshow(test_labels[i].cpu().numpy(), cmap='gray')
    
    plt.subplot(1, 3, 3)
    plt.title('Prediction')
    plt.imshow(predictions[i], cmap='gray')
    
    plt.show()

# from d2l import torch as d2l
# from tqdm import tqdm
# import pandas as pd
# #损失函数选用多分类交叉熵损失函数
# lossf = nn.CrossEntropyLoss(ignore_index=255)
# #选用adam优化器来训练
# optimizer = optim.SGD(model.parameters(),lr=0.1)
# scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=150, gamma=0.1, last_epoch=-1)
 
# #训练50轮
# epochs_num = 150
# def train_ch13(net, train_iter, test_iter, loss, trainer, num_epochs,scheduler,
#                devices=d2l.try_all_gpus()):
#     timer, num_batches = d2l.Timer(), len(train_iter)
#     animator = d2l.Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=[0, 1],
#                             legend=['train loss', 'train acc', 'test acc'])
#     # net = nn.DataParallel(net, device_ids=devices)#.to(devices[0])
    
#     loss_list = []
#     train_acc_list = []
#     test_acc_list = []
#     epochs_list = []
#     time_list = []
#     for epoch in range(num_epochs):
#         # Sum of training loss, sum of training accuracy, no. of examples,
#         # no. of predictions
#         metric = d2l.Accumulator(4)
#         for i, (features, labels) in enumerate(train_iter):
#             timer.start()
#             l, acc = d2l.train_batch_ch13(
#                 net, features, labels.long(), loss, trainer, devices)
#             metric.add(l, acc, labels.shape[0], labels.numel())
#             timer.stop()
#             if (i + 1) % (num_batches // 5) == 0 or i == num_batches - 1:
#                 animator.add(epoch + (i + 1) / num_batches,
#                              (metric[0] / metric[2], metric[1] / metric[3],
#                               None))
#         try:
#             test_acc = d2l.evaluate_accuracy_gpu(net, test_iter)#_gpu
#             animator.add(epoch + 1, (None, None, test_acc))
            
#             print(f"epoch {epoch+1} --- loss {metric[0] / metric[2]:.3f} ---  train acc {metric[1] / metric[3]:.3f} --- test acc {test_acc:.3f} --- cost time {timer.sum()}")
#         except Exception as e:
#             print(e)
#         scheduler.step()
#         #---------保存训练数据---------------
#         # df = pd.DataFrame()
#         # loss_list.append(metric[0] / metric[2])
#         # train_acc_list.append(metric[1] / metric[3])
#         # test_acc_list.append(test_acc)
#         # epochs_list.append(epoch+1)
#         # time_list.append(timer.sum())
        
#         # df['epoch'] = epochs_list
#         # df['loss'] = loss_list
#         # df['train_acc'] = train_acc_list
#         # df['test_acc'] = test_acc_list
#         # df['time'] = time_list
#         # df.to_excel("dataset/DFAnet_camvid.xlsx")
#         #----------------保存模型-------------------
#         if np.mod(epoch+1, 5) == 0:
#             torch.save(model.state_dict(), f'dataset/DFAnet_{epoch+1}.pth')

# train_ch13(model, train_loader, val_loader, lossf, optimizer, epochs_num,scheduler)