import os
import glob
import numpy as np
import cv2
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms, models
import matplotlib.pyplot as plt
from tqdm import tqdm  # 进度条库

class CityscapesDataset(Dataset):
    def __init__(self, image_paths, label_paths, transform=None):
        self.image_paths = image_paths
        self.label_paths = label_paths
        self.transform = transform

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image = cv2.imread(self.image_paths[idx])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        label = cv2.imread(self.label_paths[idx], cv2.IMREAD_GRAYSCALE)
        
        if self.transform:
            image = self.transform(image)
        
        label = cv2.resize(label, (image.shape[2], image.shape[1]), interpolation=cv2.INTER_NEAREST)
        label = torch.from_numpy(label).long()
        
        # 确保标签在有效范围内
        label[label >= 20] = 19  # 假设19是最高的有效类索引

        return image, label

# 获取所有图像和标签的路径
image_paths = sorted(glob.glob('dataset/leftImg8bit_trainvaltest/leftImg8bit/train/*_leftImg8bit.png'))
label_paths = sorted(glob.glob('dataset/gtFine/train/*_gtFine_labelIds.png'))

# 确保路径数量一致
assert len(image_paths) == len(label_paths)

# 数据增强和预处理
transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.Resize((256, 512)),
    transforms.ToTensor(),
])

dataset = CityscapesDataset(image_paths, label_paths, transform=transform)
dataloader = DataLoader(dataset, batch_size=4, shuffle=True)

# 验证集加载
val_image_paths = sorted(glob.glob('dataset/leftImg8bit_trainvaltest/leftImg8bit/val/*_leftImg8bit.png'))
val_label_paths = sorted(glob.glob('dataset/gtFine/val/*_gtFine_labelIds.png'))
val_dataset = CityscapesDataset(val_image_paths, val_label_paths, transform=transform)
val_dataloader = DataLoader(val_dataset, batch_size=4, shuffle=False)




# 使用PyTorch中的预训练DeepLabV3模型
model = models.segmentation.deeplabv3_resnet101(pretrained=False)
model.classifier[4] = nn.Conv2d(256, 20, kernel_size=(1, 1), stride=(1, 1))  # Cityscapes有19类+背景，共20类
# model = model.cuda()

criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(), lr=0.001)

# 训练模型
num_epochs = 10
for epoch in range(num_epochs):
    model.train()
    running_loss = 0.0
    with tqdm(total=len(dataloader), desc=f"Epoch {epoch+1}/{num_epochs}", unit="batch") as pbar:
        for images, labels in dataloader:
            # images = images.cuda()
            # labels = labels.cuda()

            optimizer.zero_grad()
            outputs = model(images)['out']
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
            
            running_loss += loss.item()
            pbar.set_postfix(loss=running_loss/(pbar.n + 1))
            pbar.update(1)
    
    print(f"Epoch [{epoch+1}/{num_epochs}], Loss: {running_loss/len(dataloader)}")

    # 验证模型
    model.eval()
    val_loss = 0.0
    with torch.no_grad():
        for images, labels in val_dataloader:
            # images = images.cuda()
            # labels = labels.cuda()
            
            outputs = model(images)['out']
            loss = criterion(outputs, labels)
            val_loss += loss.item()
    
    print(f"Validation Loss: {val_loss/len(val_dataloader)}")


# 选择一些测试图像进行预测
test_images, test_labels = next(iter(val_dataloader))
# test_images = test_images.cuda()
# test_labels = test_labels.cuda()

# 进行预测
model.eval()
with torch.no_grad():
    predictions = model(test_images)['out']
predictions = torch.argmax(predictions, dim=1).cpu().numpy()

# 将预测结果转换为类别标签并可视化
for i in range(4):  # 批量大小为4
    plt.figure(figsize=(15, 5))
    
    plt.subplot(1, 3, 1)
    plt.title('Original Image')
    plt.imshow(test_images[i].permute(1, 2, 0).cpu().numpy())
    
    plt.subplot(1, 3, 2)
    plt.title('Ground Truth')
    plt.imshow(test_labels[i].cpu().numpy(), cmap='gray')
    
    plt.subplot(1, 3, 3)
    plt.title('Prediction')
    plt.imshow(predictions[i], cmap='gray')
    
    plt.show()
